class T2Addr < ExpressionType
	attr_accessor :ip, :bp
	def initialize(ip, bp)
		@ip, @bp = ip, bp
	end

	def to_s; Expression[@ip].to_s.rjust(6) + '_' + Expression[@bp].to_s.ljust(6) end

	def hash; @ip.hash ^ @bp.hash end
	def eql?(o) o.class == self.class and o.ip == @ip and o.bp == @bp end
	alias == eql?
	def inspect; "T2Addr.new(#{@ip.inspect}, #{@bp.inspect})" end
	def externals; Expression[@ip, :+, @bp].externals end
	def reduce_rec
		a = T2Addr.new(Expression[@ip].reduce, Expression[@bp].reduce)
		a = Expression::Unknown if a.ip == Expression::Unknown or a.bp == Expression::Unknown
		a
	end
	def bind(h) T2Addr.new(Expression[@ip].bind(h), Expression[@bp].bind(h)) end
	def complexity; Expression[@ip, :+, @bp].complexity end
	def expr_indirections; Expression[@ip, :+, @bp].expr_indirections end
	def -(o) 0 end
	
	def self.[](i, b) new(Expression[i], Expression[b]) end
end

class T2DecodedInstruction < DecodedInstruction
	attr_accessor :address, :next_addr
	def initialize(cpu, addr)
		@address = addr
		@instruction = Instruction.new cpu
		@bin_length = 1
	end

	def show
		s = "    #{@instruction}"
		s = s.ljust(36) + " ;  @#{@address}  #{@comment.join(' ') if comment}"
		s
	end
	def to_s; "#{@address} #{@instruction}" end
end

class T2CPU < CPU
	class Handler < Opcode
		attr_accessor :name, :key_args, :key_hand, :args, :binding, :props
		def initialize(name, key_args, key_hand, args, binding)
			@name, @key_args, @key_hand, @args, @binding = name, key_args, key_hand, args, binding
			@props = {}
			@props[:stopexec] = true if not key_hand
		end
	end

	# hash addr => Handler
	attr_accessor :handler, :ia32dasm

	def initialize(pe)
		super()

		@ia32dasm = pe.init_disassembler
		df = @ia32dasm.function[:default] = @ia32dasm.cpu.disassembler_default_func
		df.btbind_callback = nil
		df.btfor_callback = proc { |*a|
			if di = a[0].decoded[a[3]] and di.opcode.name == 'call': a[1]
			else []
			end
		}
		df.backtrace_binding = [:eax, :ecx, :edx, :esp].inject({}) { |h, r| h.update r => Expression::Unknown }

		@handler = {}

		File.read('saved-handlers').each { |a|
			a = eval(a)
			@handler[a[0]] = Handler.new(a[1], a[2], a[3], a[4], a[5])
		} if File.exist? 'saved-handlers'
	end

	REGNAMES = {
		:r0 => :esp_init,
		:r4 => :esp,
		:r8 => :ebp,
		:rc => :host_esp,
		:r14 => :host_data,
		:r18 => :retval,
		:r1c => :ctx_256,
		:r20 => :zero_20,
		:r2c => :base_64_2c,
	}

	def decode_instruction(foo, addr)
		return if not @ia32dasm.get_section_at(addr.ip)
		h = @handler[addr.ip] ||= disassemble_handler(addr.ip)
		@ia32dasm.auto_label_at(addr.bp, 'vminstr')
		return if not h or not s = @ia32dasm.get_section_at(addr.bp+8)
		di = T2DecodedInstruction.new self, addr
		di.opcode = h
		di.instruction.opname = h.name

		rdimm = proc {
			@ia32dasm.add_xref(s[1]+s[0].ptr, Xref.new(:r, addr.ip, 4))
			i = s[0].decode_imm(:u32, @ia32dasm.cpu.endianness) ^ (h.key_args || 0x10000_0000)
			i = i-(1<<32) if (i>>20) == 0xfff and h.key_args
			i
		}
		rdreg = proc { r = "r#{rdimm[].to_s(16)}".to_sym ; REGNAMES.fetch(r, r) }

		h.args.each { |a|
			di.instruction.args <<
			case a
			when :reg: rdreg[]
			when :regind: Indirection[rdreg[], 4, di.address]
			when :imm: Expression[rdimm[]]
			when :imm1: Expression[1]
			when :regindb: Indirection[rdreg[], 1, di.address]
			when :label: T2Addr.new(rdimm[]+0x13000, rdimm[]+0x13000)
			when :jmpind
				s[0].ptr -= 8
				i1 = s[0].decode_imm(:u32, @ia32dasm.cpu.endianness) ^ h.binding[:jmpkey_eip]
				r1 = "r#{i1.to_s(16)}".to_sym
				r1 = REGNAMES.fetch(r1, r1)
				i2 = s[0].decode_imm(:u32, @ia32dasm.cpu.endianness) ^ h.binding[:jmpkey_ebp]
				r2 = "r#{i2.to_s(16)}".to_sym
				r2 = REGNAMES.fetch(r2, r2)
				T2Addr.new(Expression[r1, :+, 0x13000], Expression[r2, :+, 0x13000])
			end
		}

		if h.key_hand and h.key_args
			@ia32dasm.add_xref(addr.bp,   Xref.new(:r, addr.ip, 4))
			@ia32dasm.add_xref(addr.bp+4, Xref.new(:r, addr.ip, 4))
			di.next_addr = T2Addr.new(
				solve_ind(Expression[[Indirection[ addr.bp,         4], :^, h.key_hand], :+, 0x13000]),
				solve_ind(Expression[[Indirection[[addr.bp, :+, 4], 4], :^, h.key_args], :+, 0x13000])
			)
		end

		di
	end

	def get_xrefs_x(dasm, di)
		di.instruction.args.grep(T2Addr)
	end
	def backtrace_binding(di)
		a0 = di.instruction.args[0]
		a1 = di.instruction.args[1]
		a2 = di.instruction.args[2]
		a3 = di.instruction.args[3]
		case di.instruction.opname
		when 'mov': {a0 => Expression[a1]}
		when 'add': {a0 => Expression[a0, :+, a1]}
		when 'xor': {a0 => Expression[a0, :^, a1]}
		when 'and': {a0 => Expression[a0, :&, a1]}
		when 'shr': {a0 => Expression[a0, :>>, a1]}
		when 'mul': {a0 => Expression[[a0, :*, a1], :&, 0xffff_ffff]}
		when 'xorb1': {a0 => Expression[[a0, :&, 1], :^, 1]}
		when 'seteql', 'setgt', 'setlt': {a0 => Expression::Unknown}
		when 'syscall_alloc_ptr': {a0 => Expression::Unknown}
		when 'syscall_free_exitvm': {}
		when 'syscall_free': {}
		when 'syscall_md5': {}
		when 'decryptcopy': {Indirection[a0, 4] => Expression::Unknown} #Indirection[a1+0x13000, 4]}
		when 'host_call': {a0 => Expression::Unknown}
		when 'nop', 'trap', 'jrz', 'jmp', 'host_push': {}
		when /^unk /: puts "Unk instr bt: #{di.to_s[0, 17]}.." ; {}
		else puts "Unknown instruction to backtrace: #{di}"#, caller, di.opcode.binding.inspect
			{}
		end
	end

	def replace_instr_arg_immediate(i, expr, n)
		expr = Expression[expr].reduce_rec
		n    = Expression[n   ].reduce_rec
		i.args.map! { |e| e == expr ? n : e }
	end

	def handler_binding(handler_addr)
		l = @ia32dasm.auto_label_at(handler_addr, 'handler', 'loc')

		@ia32dasm.cpu.dasm ||= @ia32dasm	rescue nil # allow the cpu to disasm the other rdtsc branch
		# disassemble the handler code
		oldv = $VERBOSE
		$VERBOSE = false
		@ia32dasm.disassemble(handler_addr)
		$VERBOSE = oldv

		bd32 = [:eax, :ebx, :ecx, :edx, :ebp, :edi, :esi].inject({}) { |h, r| h.update Expression[r, :&, 0xffff_ffff] => Expression[r] }
		sim32 = proc { |e|
			case e
			when ::Array: e.map { |ee| sim32[ee] }
			when Indirection, Expression: e.bind(bd32)
			else e
			end
		}

		h_bind = {}
		add_bt = proc { |key, vals|
			key = sim32[key]
			vals = sim32[vals]
			next if key.kind_of? Indirection and key.target.externals == [:ebx] and o = Expression[key.target, :-, :ebx].reduce and o.kind_of? Integer and o >= 0x100 and o < 66000
			next if vals.empty?
			next if vals == [h_bind[key]]
			case h_bind[key]
			when nil: h_bind[key] = (vals.length == 1) ? vals.first : vals
			when ::Array: h_bind[key] |= vals
			else h_bind[key] = [h_bind[key]] | vals
			end
		}

		blocks = [handler_addr]
		blocks_done = [Expression::Unknown]

		# walk the handler code searching for xrefs
		while b = blocks.pop
			b = @ia32dasm.normalize b
			next if blocks_done.include? b or not @ia32dasm.decoded[b]
			blocks_done << b
			b = @ia32dasm.decoded[b].block

			b.each_to { |t, type| blocks << t }

			di = b.list.last
			@ia32dasm.cpu.get_xrefs_x(self, di).to_a.each { |e|
				hend = false
				@ia32dasm.backtrace(e, di.address, :snapshot_addr => handler_addr).each { |e|
					if di.opcode.name == 'call' and b.to_subfuncret # API call
						add_bt['call', [e]]
						(0..4).each { |i|
							a = @ia32dasm.backtrace(Indirection[[:esp, :+,  4*i], 4], di.address, :snapshot_addr => handler_addr)
							a.delete_if { |aa| aa.rexpr.kind_of? Indirection and aa.rexpr.target.externals.include? :esp }
							add_bt["call_arg#{i}", a] if not a.empty?
						}
						next
					end

					next if @ia32dasm.normalize(e).kind_of?(::Integer) # in-handler jmp

					# end of handler
					add_bt[:eip, [e]]
					next if hend
					hend = true
					[:ebx, :esp, :ebp].each { |r|
						add_bt[r, @ia32dasm.backtrace(Expression[r], di.address,
							:include_start => true, :snapshot_addr => handler_addr)]
					}
				}
			}

			done_wptr = []
			# for each :w xref, trace the target
			# reverse_each to have only one xref with  mov [bla], 42 ; add [bla], 28
			# XXX mov [bla], 42 ; jmp foo ; foo: add [bla], 28
			b.list.reverse_each { |di|
				@ia32dasm.cpu.get_xrefs_rw(self, di).to_a.each { |type, e, len|
					next if type != :w or e.externals.include? :esp
					val = @ia32dasm.backtrace(Indirection[e, len, di.address], di.address,
							:include_start => true, :snapshot_addr => handler_addr)
					ptr = @ia32dasm.backtrace(e, di.address, :snapshot_addr => handler_addr)
					e = Expression[ptr.first] if ptr and ptr.length == 1
					if m = Expression[e].match(Expression['base', :+, [[4, :*, 'idx'], :+, 'imm']], 'base', 'idx', 'imm') ||
					       Expression[e].match(Expression['base', :+, [4, :*, 'idx']], 'base', 'idx')
						mb = @ia32dasm.backtrace(m['base'], di.address, :snapshot_addr => handler_addr)
						e = Expression[Expression[e].bind(m['base'] => Expression[mb.first]).reduce] if mb.length == 1
					end
					next if done_wptr.include? e
					done_wptr << e
					add_bt[Indirection[e, len, di.address], val]
				}
			}
		end

		h_bind.delete_if { |k, v| Expression[k].reduce == Expression::Unknown }
		h_bind
	end

	def disassemble_handler(handler_addr)
		h_bind = handler_binding(handler_addr)

		h_bind.delete :esp if h_bind[:esp] == Expression[:esp]
		h_bind.delete :ebx if h_bind[:ebx] == Expression[:ebx]

		if not h_bind[:ebx] and h_bind[:ebp]
			sec_base = Indirection[[:ebx, :+, 0x14], 4]

			nebp = h_bind[:ebp]
			nebp = [nebp] if nebp.kind_of? Expression
			keys_args = nebp.map { |e| Expression[[e, :-, sec_base], :^, Indirection[[:ebp, :+, 4], 4]].reduce }.grep(::Integer).uniq
			if keys_args.length == 1
				key_args = keys_args.first
				if h_bind[:ebp].kind_of? ::Array
					h_bind[:ebp].delete_if { |e| Expression[[e, :-, sec_base], :^, Indirection[[:ebp, :+, 4], 4]].reduce.kind_of? ::Integer }
					case h_bind[:ebp].length
					when 1: h_bind[:ebp] = h_bind[:ebp].first
					when 0: h_bind.delete :ebp
					end
				else h_bind.delete :ebp
				end

				b = {}
				(0..4).each { |i|
					e = Expression[Indirection[[:ebp, :+, 8+4*i], 4], :^, key_args]
					b[e] = Expression["arg#{i}"]
					b[Expression[e, :&, 0xffff_ffff]] = Expression["arg#{i}"]
					b[Expression[Indirection[[:ebx, :+, e], 4], :&, 0xffff_ffff]] = Expression["reg#{i}"]
					b[Indirection[[:ebx, :+, e], 4]] = Expression["reg#{i}"]
					b[Indirection[[:ebx, :+, e], 1]] = Expression["reg#{i}b"]
				}
				h_bind.keys.dup.each { |k|
					case h_bind[k] # reduce values
					when Expression: h_bind[k] = Expression[h_bind[k].bind(b).reduce]
					when ::Array: h_bind[k] = h_bind[k].map { |e| e.kind_of?(Expression) ? Expression[e.bind(b).reduce] : e }
					end
					case k # reduce keys
					when Expression, Indirection: h_bind[k.bind(b).reduce_rec] = h_bind.delete(k)
					end
				}
			end

			neip = h_bind[:eip]
			neip = [neip] if neip.kind_of? Expression
			keys_hand = neip.map { |e| Expression[[e, :-, sec_base], :^, Indirection[:ebp, 4]].reduce }.grep(::Integer).uniq
			if keys_hand.length == 1 and key_args
				key_hand = keys_hand.first
				if h_bind[:eip].kind_of? ::Array
					h_bind[:eip].delete_if { |e| Expression[[e, :-, sec_base], :^, Indirection[:ebp, 4]].reduce.kind_of? ::Integer }
					case h_bind[:eip].length
					when 1: h_bind[:eip] = h_bind[:eip].first
					when 0: h_bind.delete :eip
					end
				else h_bind.delete :eip
				end
			end
		end

		blockcnt = 0
		@ia32dasm.decoded[handler_addr].block.each_to { |t, r| blockcnt += 1 if @ia32dasm.decoded[@ia32dasm.normalize(t)] }

		if blockcnt == 0
			case h_bind
			when {}: name = 'nop' ; args = []
			when {'reg0' => Expression['arg1']}: name = 'mov' ; args = [:reg, :imm]
			when {'reg0' => Expression['reg1']}: name = 'mov' ; args = [:reg, :reg]
			when {'reg0' => Expression[Indirection['reg1', 4]]}: name = 'mov' ; args = [:reg, :regind]
			when {Indirection['reg0', 4] => Expression['reg1']}: name = 'mov' ; args = [:regind, :reg]
			when {'reg0' => Expression[Indirection['reg1', 4], :&, 0xffff_ffff]}: name = 'mov' ; args = [:reg, :regind]
			when {Indirection['reg0', 4] => Expression['reg1']}: name = 'mov' ; args = [:regind, :reg]
			when {'reg0' => Expression['reg0', :+, 'reg1']}: name = 'add' ; args = [:reg, :reg]
			when {'reg0' => Expression['reg0', :^, 'reg1']}: name = 'xor' ; args = [:reg, :reg]
			when {'reg0' => Expression['reg0', :&, 'reg1']}: name = 'and' ; args = [:reg, :reg]
			when {'reg0' => Expression[['reg1', :*, 'reg0'], :&, 0xffff_ffff]}: name = 'mul' ; args = [:reg, :reg]
			when {'reg0' => Expression['reg0', :>>, ['reg1', :%, 32]]}: name = 'shr' ; args = [:reg, :reg]
			when {'reg0' => Expression['reg0', :^, 1]}: name = 'xor' ; args = [:reg, :imm1]
			when {'reg0b' => Expression['reg0b', :^, 'reg1b']}: name = 'xor' ; args = [:reg, :reg]
			when {Indirection['reg0', 1] => Expression['reg1b']}: name = 'mov' ; args = [:regindb, :reg]	# regb
			when {'reg0b' => Expression[Indirection['reg1', 1]]}: name = 'mov' ; args = [:reg, :regindb]	# regb
			when {'reg0b' => Expression[['reg0b', :&, 1], :^, 1]}: name = 'xorb1' ; args = [:reg]	# regb
			when {:esp => Expression[:esp, :+, -4]}: name = 'host_push' ; args = [:reg]
			else
				if h_bind[:eip] and h_bind[:ebp] and h_bind.length == 2 and
					mi = h_bind[:eip].match(Expression[Indirection[[:ebx, :+, [Indirection[ :ebp,         4], :^, 'key']], 4], :+, Indirection[[:ebx, :+, 0x14], 4]], 'key') and
					mp = h_bind[:ebp].match(Expression[Indirection[[:ebx, :+, [Indirection[[:ebp, :+, 4], 4], :^, 'key']], 4], :+, Indirection[[:ebx, :+, 0x14], 4]], 'key')
					name = 'jmp'; args = [:jmpind] ; h_bind.update nh_bind = {:jmpkey_eip => mi['key'], :jmpkey_ebp => mp['key'] }
				else name = 'unk single ' + h_bind.inspect ; args = [:imm, :imm, :imm, :imm]
				end
			end
		elsif h_bind == {}: name = 'trap' ; args = []
		elsif h_bind == {:eip => Expression['arg0', :+, Indirection[[:ebx, :+, 0x14], 4]],
			         :ebp => Expression['arg1', :+, Indirection[[:ebx, :+, 0x14], 4]]} and
				@ia32dasm.decoded[handler_addr].block.list.last.opcode.name == 'jz'
			name = 'jrz' ; args = [:label, :reg]
		elsif h_bind == {'reg0' => [Expression[0], Expression[1]]} and
				@ia32dasm.decoded[handler_addr].block.list.last.opcode.name == 'jnz'
			name = 'seteql' ; args = [:reg, :reg]
		elsif h_bind == {'reg0' => [Expression[0], Expression[1]]} and
				@ia32dasm.decoded[handler_addr].block.list.last.opcode.name == 'jle'
			cmp = @ia32dasm.decoded[handler_addr].block.list.reverse.find { |di| di.opcode.name == 'cmp' }
			if cmp.instruction.args.first.kind_of? Ia32::Reg
				name = 'setlt'
			else
				name = 'setgt'
			end
			args = [:reg, :reg]
		elsif h_bind == {'call' => Expression['reg0'], 'reg0' => Expression::Unknown, :esp => Expression::Unknown}
			name = 'host_call' ; args = [:reg]
		elsif c = h_bind['call'] and c = c.reduce_rec and c.kind_of? Indirection and
				idx = Expression[c.target, :-, Indirection[[:ebx, :+, 0x14], 4]].reduce and idx.kind_of? ::Integer
			h_bind.delete :esp
			bd = h_bind.dup
			bd.delete 'call'
			args = []
			name = 'syscall_'
			case idx
			when 0: name << 'alloc'
				if bd['call_arg0'] == Expression['reg1'] and bd['reg0'] == Expression::Unknown and bd.length == 2
					name << '_ptr'
					args = [:reg, :reg]
					bd = {}
				end
			when 4: name << 'free'
				if bd['call_arg0'] == Expression[:ebx] and bd[:eip] == Expression::Unknown
					name << '_exitvm'
					bd = {}
				elsif bd == {'call_arg0' => Expression['reg0']}
					args = [:reg]
					bd = {}
				end
			when 12: name << '<rewritten>'
			when 16: name << 'dbgsprintf'
			when 24: name << 'kgetprocaddr'
				if s = @ia32dasm.get_section_at(bd['call_arg0']) and n = s[0].read(128) and n.index(0)
					n = n[0, n.index(0)]
					name << ?( << n.inspect << ?)
					bd.delete 'call_arg0'
					if bd.empty?
						args = []
						name = 'trap' if n == 'KdDebuggerEnabled'
					end
				end
			when 28: name << 'md5'
				if bd == {'call_arg0' => Expression['reg0'], 'call_arg1' => Expression['reg1'], 'call_arg2' => Expression['reg2']}
					args = [:reg, :reg, :reg]
					bd = {}
				end
			end
			if not bd.empty?
				args = [:imm, :imm, :imm, :imm]
				name << bd.inspect
			end
		elsif h_bind.length == 1 and blockcnt == 1 and h2 = @ia32dasm.decoded[handler_addr].block.to_normal and di2 = @ia32dasm.decoded[h2.first] and
				di2.block.list.last.instruction.opname == 'loop' and h_bind.values.first.kind_of?(::Array) and
				v = h_bind.values.first.sort_by { |v| Expression[v].complexity }.first and
				Expression[h_bind.keys.first].match(Expression[Indirection[['reg0', :+, [[4, :*, :ecx], :+, -4]], 4]]) and
				Expression[v].match(Expression[Indirection[['arg1', :+, [Indirection[[:ebx, :+, 0x14], 4], :+, [[4, :*, 'arg3'], :+, -4]]], 4], :^, 'arg2'])
			# dest(reg), src(ctx offset), xor key, counter (rol key, counter each turn)
			name = 'decryptcopy' ; args = [:reg, :imm, :imm, :imm]
		else
			name = 'unk ' << h_bind.inspect ; args = [:imm, :imm, :imm, :imm]
		end

		h = Handler.new(name, key_args, key_hand, args, h_bind)

		c = @ia32dasm.comment[handler_addr] ||= []
		c << "handler type: #{h.name} #{h.args.join(', ')}" if h.name[0, 4] != 'unk '
		h.binding.each { |k, v| c << "#{k.inspect} <- #{v.inspect}" }

		File.open('saved-handlers', 'a') { |fd|
			fd.puts '[' + [handler_addr, name, key_args, key_hand, args, nh_bind || {}].map { |a| a.kind_of?(::Integer) ? ('0x%x' % a) : a.inspect }.join(', ') + ']'
		} if not name.index ?{
		
		puts "#{h.name} #{h.args.join(', ')}"

		h
	end

	def solve_ind(i)
		case i
		when Indirection
			p = solve_ind(i.target)
			if s = @ia32dasm.get_section_at(p)
				s[0].decode_imm("a#{i.len*8}".to_sym, @ia32dasm.cpu.endianness)
			else
				Expression::Unknown
			end
		when Expression: i.bind(i.expr_indirections.inject({}) { |b, e| b.update e => Expression[solve_ind(e)] }).reduce
		when :unknown: i
		else Expression::Unknown # raise i.inspect
		end
	end
end

class T2Disassembler < Disassembler
	def normalize(a)
		aa = Expression[a].bind(@prog_binding).reduce_rec
		return aa if aa.kind_of? T2Addr
		super
	end

	def auto_label_at(addr, base='xref', *rewritepfx)
		if a = Expression[addr].reduce_rec and a.kind_of? T2Addr
			addrstr = "#{base}_#{a}"
			if not l = @prog_binding.index(a)
				l = @program.new_label(addrstr)
				@prog_binding[l] = a
			elsif rewritepfx.find { |p| base != p and addrstr.sub(base, p) == l }
				newl = @program.new_label(addrstr)
				rename_label l, newl
				l = newl
			end
			l
		else
			super
		end
	end

	def get_section_at(a)
		a = normalize(a)
		return if not a.kind_of? T2Addr or not @cpu.ia32dasm.get_section_at(a.ip) || @cpu.ia32dasm.get_section_at(a.bp)
		e = @edata ||= EncodedData.new
		e.ptr = a.ip ^ a.bp
		[e, a]
	end
end
